package trivial;

import static common.Log.COMM;
import static common.Log.TEST;
import static common.Log.LOG_ON;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.net.StandardSocketOptions;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.nio.channels.ServerSocketChannel;
import java.nio.channels.SocketChannel;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import org.apache.log4j.Level;

import common.FullDuplexMsgWorker;
import common.ReadMessageStatus;
import https.HttpsProtocol;
import json.TrivialJson;
import json.TrivialQuestion;

/**
 * 
 * @author Éric Lallet.
 *
 */
public class TrivialServer {
	/**
	 * number of seconds before sending the correct answer to the players.
	 */
	private static final int  WAITINGTIMER = 30; // seconds
	/**
	 * number of seconds before sending new question after sending the  answer of the precedent one.
	 */
	private static final int PENDINGTIMER = 30; // seconds
	/**
	 * number of seconds before stopping server if no connection happens.
	 */
	private static final int SLEEPINGTIMER = 3600; // seconds (1 hour)  
	/**
	 * number of milliseconds in one second.
	 */
	private static final int THOUSAND = 1000;
	
	/**
	 * server port number. 
	 */
	private int port;
	
	/**
	 * URL of the trivial json service.
	 */
	private String trivialHttpUrl;
	
	/**
	 * Association between the channel and the worker.
	 */
	private	Map<SelectionKey, FullDuplexMsgWorker> theClients;
	
	/**
	 * Association between received answers and players.
	 */
	private Map<SelectionKey, Boolean> hasAnswered;
	
	/**
	 * 
	 */
	private Selector selector;
	
	
	
	/**
	 * state of this server.
	 */
	private TrivialServerState serverState = TrivialServerState.SLEEPING;  
	
	
	/**
	 * TrivialQuestion with no question.
	 */
	private  final TrivialQuestion voidTrivialQuestion = new TrivialJson().getTrivialQuestion();
	
	/**
	 * TrivialQuestion to be sent to the clients.
	 */
	private TrivialQuestion trivialQuestion;
	
	/**
	 * players answer list + correct answer.
	 */
	private  AnswersList answersList = null;
	
	
	/**
	 * the beginning time of the sleeping state.
	 */
	private long sleepingTime;
	/**
	 * the beginning time of the waiting state.
	 */
	private  long waitingTime;
	/**
	 * the beginning time of the pending state.
	 */
	private long pendingTime;
	
	

	
	/**
	 * 
	 * @param port
	 *            server port number.
	 * @param trivialHttpUrl
	 *            URL of the trivial json service.
	 * @throws IOException 
	 *            the exception thrown in case of IO problem.
	 */
	public TrivialServer(final int port, final String trivialHttpUrl) throws IOException {
		this.port = port;
		this.trivialHttpUrl = trivialHttpUrl;
		this.play();
	}
	
	
	
	
	/**
	 * turn to false all value of hasAnswered Map. 
	 */
	private void resetHasAnswered() {
		for (SelectionKey key: hasAnswered.keySet()) {
			hasAnswered.put(key, false);
		}
	}
	
	/**
	 * test if the is still client who has not sent any answer for the current question.
	 * @return
	 *         true if all clients have sent an answer.
	 */
	private boolean allClientsHaveAnswered() {
		for (SelectionKey key: hasAnswered.keySet()) {
			if (!hasAnswered.get(key)) {
				return false;
			}
		}
		return true;
	}
	
	/**
	 * create a new ServerSocketChannel.
	 * @return
	 *        the new ServerSocketChannel
	 * @throws IOException
	 *         the exception thrown in case of IO problem.
	 */
	private ServerSocketChannel openListenChannel() throws IOException {
		ServerSocketChannel listenChannel;
		
		// création canal socket pour le listen
		listenChannel = ServerSocketChannel.open();
		listenChannel.setOption(StandardSocketOptions.SO_REUSEADDR, true);
		listenChannel.bind(new InetSocketAddress(port));
		
		// passage en mode assynchrone
		listenChannel.configureBlocking(false);
		
		return listenChannel;
		
	}
	
	/**
	 * accept a new client.
	 * @param listenChannel
	 *                listenChannel for the accept.
	 * @throws IOException
	 *                    the exception thrown in case of IO problem.
	 */
	private void acceptNewPlayer(final ServerSocketChannel listenChannel) throws IOException {
		SocketChannel rwChan;
		SelectionKey newKey;
		FullDuplexMsgWorker worker;

		// un nouveau canal pour le nouveau client
		rwChan = listenChannel.accept();
		if (rwChan != null) {
			rwChan.configureBlocking(false);

			// le canal du client est surveillé pour des lectures.
			newKey = rwChan.register(selector, SelectionKey.OP_READ);

			// création du worker pour ce client. Ajout du client à la liste.
			worker = new FullDuplexMsgWorker(rwChan);
			theClients.put(newKey, worker);
			hasAnswered.put(newKey, false);
		// more services: send the current question to new player
			if (theClients.size() > 1) {
				worker.sendMsg(MessType.TRIVIALQUESTIONTYPE, trivialQuestion);
			}

		}

	}
	
	/**
	 * Ask a new trivial card to the http server, and send the question to all connected players.
	 * @throws IOException
	 *                    the exception thrown in case of IO problem.
	 */
	private void sendNewQuestion() throws IOException {
		HttpsProtocol  httpsProtocol = new HttpsProtocol();
		httpsProtocol.connectTo(trivialHttpUrl);
		String jsonData = httpsProtocol.recvData();
		TrivialJson trivialJson = new TrivialJson(jsonData);
		
		trivialQuestion = trivialJson.getTrivialQuestion();
		for (FullDuplexMsgWorker worker: theClients.values()) {
			worker.sendMsg(MessType.TRIVIALQUESTIONTYPE, trivialQuestion);
		
		}
		answersList = new AnswersList();
		answersList.setCorrectAnswer(trivialQuestion.getCorrectAnswerNum());
		resetHasAnswered();
	}
		

	
	/**
	 * receive an answer from player. Add this answer in the server annsersList.
	 * @param key
	 *           key bind to client channel in theClients Map.
	 * @throws IOException
	 *           the exception thrown in case of IO problem.
	 */
	private void recieveAnswerFromPlayer(final SelectionKey key) throws IOException {
		FullDuplexMsgWorker worker;
		ReadMessageStatus status;
		
		// cet évènement est une lecture à faire
		// on retrouve le client concerné par sa clef.
		worker = theClients.get(key);
		
		
		// on avance la lecture du message de ce client.
		status = worker.readMessage();
		if (status == ReadMessageStatus.ChannelClosed) { // closed channel
			// le client est parti, on le retire de la liste.
			theClients.remove(key);
			hasAnswered.remove(key);
			key.cancel();
			worker.close();
		}
		if (status == ReadMessageStatus.ReadDataCompleted) {
				AnswersList oneAnwser = (AnswersList) worker.getData();
				answersList.add(oneAnwser.getNickname(0), oneAnwser.getAnswerNum(0));
				hasAnswered.put(key, true);
			
		}
		if (LOG_ON && COMM.isEnabledFor(Level.DEBUG)) {
			COMM.debug("Received message status:" + status);
		}
	}
	
	/**
	 * send to all players the players answer list with the correct answer.
	 * @throws IOException
	 *                   the exception thrown in case of IO problem.
	 */
	private void sendCorrectAnswer() throws IOException {
		
		
		
		for (FullDuplexMsgWorker worker: theClients.values()) {
			worker.sendMsg(MessType.ANSWERSLISTTYPE, answersList);
		}
		
		
	}
	
	
	/**
	 * change  the server state and send data to the client if needed.
	 * @return
	 *        timer of the next select().
	 * @throws IOException
	 *          the exception thrown in case of IO problem.
	 */
	private long changeStateAndSendData() throws IOException {
		long now = System.currentTimeMillis();
		int numberOfClients = theClients.size();
		long selectTimer = 0;
		
		
		switch (serverState) {
		case SLEEPING:
			selectTimer = sleepingTime + (SLEEPINGTIMER * THOUSAND) - now;
			if (numberOfClients > 0) {
				sendNewQuestion();
				serverState = TrivialServerState.WAITING;
				waitingTime =  System.currentTimeMillis();
				return WAITINGTIMER * THOUSAND;
			}
			if (selectTimer <= 0) {
				serverState = TrivialServerState.FINALSTATE;
				return 0; // select not reached, this value is never used. 
			}
			break;
		case WAITING:
			selectTimer = waitingTime + (WAITINGTIMER * THOUSAND) - now;
			if (numberOfClients == 0) {
				serverState = TrivialServerState.SLEEPING;
				sleepingTime = System.currentTimeMillis();
				return SLEEPINGTIMER * THOUSAND;
				
			}
			if ((selectTimer <= 0) || (allClientsHaveAnswered())) {
				sendCorrectAnswer();
				serverState = TrivialServerState.PENDING;
				pendingTime =  System.currentTimeMillis();
				trivialQuestion = voidTrivialQuestion; // this "no" question is sent to the new client
				                                       // when the server is in PENDING state.
				return PENDINGTIMER * THOUSAND;
			}
			break;
		case PENDING:
			selectTimer = pendingTime + (PENDINGTIMER * THOUSAND) - now;
			if (numberOfClients == 0) {
				serverState = TrivialServerState.SLEEPING;
				sleepingTime = System.currentTimeMillis();
				return SLEEPINGTIMER * THOUSAND;
				
			}
			if (selectTimer <= 0) {
				sendNewQuestion();
				serverState = TrivialServerState.WAITING;
				waitingTime =  System.currentTimeMillis();
				return WAITINGTIMER * THOUSAND;
			}
			default:
				
		}
		return selectTimer;
	}
	

	/**
	 * loop of the server.
	 * @throws IOException
	 *         the exception thrown in case of IO problem.
	 */
	private void play() throws IOException {
		ServerSocketChannel listenChannel;
		
		
		theClients = new HashMap<SelectionKey, FullDuplexMsgWorker>();
		hasAnswered = new HashMap<SelectionKey, Boolean>();
		serverState =  TrivialServerState.SLEEPING;
		sleepingTime = System.currentTimeMillis();
		selector = Selector.open();

		listenChannel = openListenChannel();
		listenChannel.register(selector, SelectionKey.OP_ACCEPT);
		
		while (true) {
			long selectTimer = changeStateAndSendData();
			if (LOG_ON && COMM.isEnabledFor(Level.TRACE)) {
				COMM.trace("select: timer = " + selectTimer + " state = " + serverState);
			}
			
			if (serverState == TrivialServerState.FINALSTATE) {
				break;
			}

			if (LOG_ON && COMM.isEnabledFor(Level.DEBUG)) {
				COMM.debug("Total Nb keys:" + selector.keys().size());
			}
			
			selector.select(selectTimer);
			
			Set<SelectionKey> readyKeys = selector.selectedKeys();
			
			if (LOG_ON && COMM.isEnabledFor(Level.DEBUG)) {
				COMM.debug("Nb Ready keys:" + readyKeys.size());
			}
			// pacours de tous les évènements
			for (SelectionKey curKey: readyKeys) {
				if (curKey.isAcceptable()) {
					 acceptNewPlayer(listenChannel);
				} else {
					if (curKey.isReadable()) {
						recieveAnswerFromPlayer(curKey);
					}
				}
			}
			
			
		}

	}
	
	/****************************************** 
	**         methodes for testing          **                 
	**           DO NOT MODIFY               **
	*******************************************/	
	
	/**
	 * true if constructed for testing.
	 */
	private boolean testServer = false;
	
	/**
	 * don't use it. Only for tests.
	 */
	private ServerSocketChannel testListenChannel;
	

	
	/**
	 * Special constructor for unit tests.
	 * @param port
	 *            server port number.
	 * @param trivialHttpUrl
	 *             URL of the trivial json service.
	 * @param testId
	 *             id of the test.
	 * @throws IOException 
	 *             the exception thrown in case of IO problem.
	 */
	public TrivialServer(final int port, final String trivialHttpUrl, final TestId testId) throws IOException {
		this.port = port;
		this.trivialHttpUrl = trivialHttpUrl;
		this.testServer = true;
		if (LOG_ON && TEST.isInfoEnabled()) {
			TEST.info("TrivialServer instance created for " + testId);
		}
		
	}
	
	/**
	 * public method for unit test.
	 * @param testId
	 *              testId of the unit test.
	 * @throws IOException 
	 *              the exception thrown in case of IO problem.
	 */
	public void playTest(final TestId testId) throws IOException {
		if (!testServer) {
			return;
		}
		switch (testId) {
		case OPENLISTENCHANNEL:
			testListenChannel = this.openListenChannel();
			theClients = new HashMap<SelectionKey, FullDuplexMsgWorker>();
			hasAnswered = new HashMap<SelectionKey, Boolean>();
			serverState =  TrivialServerState.SLEEPING;
			selector = Selector.open();
			
			break;
		case ACCEPTNEWPLAYER:
			this.acceptNewPlayer(testListenChannel);
			break;
			
		case ACCEPTNEWPLAYER2:
			for (SelectionKey key: theClients.keySet()) {
				theClients.get(key).sendMsg(MessType.TRIVIALQUESTIONTYPE, voidTrivialQuestion);
			}
			break;
		case SENDNEWQUESTION:
			//this.acceptNewPlayer(testListenChannel);
			this.sendNewQuestion();
			break;
		case SENDNEWQUESTION2:
			for (SelectionKey key: theClients.keySet()) {
				theClients.get(key).sendMsg(MessType.ANSWERSLISTTYPE, answersList);
			}
			break;
			
		case RECEIVEANSWERFROMLPLAYER:
			for (SelectionKey key: theClients.keySet()) {
				this.recieveAnswerFromPlayer(key);
			
			}
			break;
			
		case RECEIVEANSWERFROMLPLAYER2:
			// send back the answers
			for (SelectionKey key: theClients.keySet()) {
				theClients.get(key).sendMsg(MessType.ANSWERSLISTTYPE, answersList);
			}
			
			break;
			
		case SENDCORRECTANSWER:
			this.sendCorrectAnswer();
			break;
			
		case CHANGESTATE1:
			this.changeStateAndSendData();
			break;
		case CHANGESTATE2:
			this.waitingTime -= WAITINGTIMER * THOUSAND;
			this.changeStateAndSendData();
			break;
		case CHANGESTATE3:
			this.pendingTime -= PENDINGTIMER * THOUSAND;
			this.changeStateAndSendData();
			break;
		case CHANGESTATE4:
		case CHANGESTATE5:
			// using closed connection: should reset theClients Map 
			for (SelectionKey key: theClients.keySet()) {
				this.recieveAnswerFromPlayer(key);
			
			}
			changeStateAndSendData(); // 0 client -> back to sleeping state
			break;
		default:
			break;
	
		
		}	
	}
	
}
